//////////////// Includes /////////////////////////////////////////////////////

#include "stdafx.h"
#include "SocksSvc.h"
#include "SocksClient.h"


//////////////// Classes //////////////////////////////////////////////////////

class CSocksApp : public CWinApp
{
protected:
//Methods
  virtual BOOL InitInstance();
};

class CMySocksClient : public CSocksClient
{
public:
//Methods
  virtual bool AllowThisIncomingConnection();
  virtual bool AllowThisOutboundConnection();
  virtual void OnEndRelay();
};

class CMySocksServerDirected : public CSocksServer<CMySocksClient, CSocksSvcDirectedThreadPoolQueue>
{
public:
//Methods
  virtual void OnError(_In_ const CString& sError)
  {
    //Just printf out the error to the console window
    _tprintf(_T("%s\n"), sError.GetString());
  }
};

class CMySocksServerIOCP : public CSocksServer<CMySocksClient, CIOCPThreadPoolQueue>
{
public:
//Methods
  virtual void OnError(_In_ const CString& sError)
  {
    //Just printf out the error to the console window
    _tprintf(_T("%s\n"), sError.GetString());
  }
};


//////////////// Macros / Defines /////////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


//The global variables
CMySocksServerDirected* g_pTheSocksServerDirected = nullptr;
CMySocksServerIOCP*     g_pTheSocksServerIOCP = nullptr;
CSocksApp               theApp;


//////////////// Implementation ///////////////////////////////////////////////

BOOL _stdcall ConsoleHandlerRoutine(_In_ DWORD /*dwCtrlType*/)
{
  _tprintf(_T("Received request to shut down the Socks Server\n"));
  if (g_pTheSocksServerIOCP != nullptr)
    g_pTheSocksServerIOCP->Stop();
  if (g_pTheSocksServerDirected != nullptr)
    g_pTheSocksServerDirected->Stop();
  return TRUE;
}


bool CMySocksClient::AllowThisIncomingConnection()
{
  //just print out the details to the console window
  CString sLine;
  sLine.Format(_T("Accepting client connection from %s\n"), m_sLanAddress.GetString());
  _tprintf(sLine);

  return true;
};

bool CMySocksClient::AllowThisOutboundConnection()
{
  //Validate our parameters
  ASSERT(m_pSocksRequest != nullptr);

  //Display the connections to the console window
  if (m_ConnectionRequest.m_bSocks5Request)
  {
    if (m_pSocksRequest->m_ClientAddress.si_family == AF_INET)
    {
      _tprintf(_T("Begining SOCKS5: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
    }
    else
    {
      ASSERT(m_pSocksRequest->m_ClientAddress.si_family == AF_INET6);
      _tprintf(_T("Begining SOCKS5: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
    }
  }
  else
  {
    ASSERT(m_pSocksRequest->m_ClientAddress.si_family == AF_INET);
    _tprintf(_T("Begining SOCKS4: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
  }


  return true;
}

void CMySocksClient::OnEndRelay()
{
  //Validate our parameters
  ASSERT(m_pSocksRequest != nullptr);

  //Display the connections to the console window
  if (m_ConnectionRequest.m_bSocks5Request)
  {
    if (m_pSocksRequest->m_ClientAddress.si_family == AF_INET)
    {
      _tprintf(_T("Ending SOCKS5: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
    }
    else
    {
      ASSERT(m_pSocksRequest->m_ClientAddress.si_family == AF_INET6);
      _tprintf(_T("Ending SOCKS5: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv6.sin6_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
    }
  }
  else
  {
    ASSERT(m_pSocksRequest->m_ClientAddress.si_family == AF_INET);
    _tprintf(_T("Ending SOCKS4: %s:%d - %s:%d\n"), m_sLanAddress.GetString(), static_cast<int>(m_pSocksRequest->m_ClientAddress.Ipv4.sin_port), m_sRemoteAddress.GetString(), static_cast<int>(m_nRemotePort));
  }
}

BOOL CSocksApp::InitInstance()
{
  //Initialise Sockets
  WSADATA wsaData;
  if (WSAStartup(MAKEWORD(2, 2), &wsaData) != 0) 
  {
    _tprintf(_T("Failed to initialise Winsock, Error:%u\n"), GetLastError());
    return FALSE;
  }

  //By default, use an ini file in the same location as the exe
  TCHAR szIni[_MAX_PATH];
  if (!GetModuleFileName(nullptr, szIni, _MAX_PATH))
  {
    _tprintf(_T("Failed to get module filename, Error:%u\n"), GetLastError());
    return FALSE;
  }
  TCHAR szDrive[_MAX_DRIVE];
  szDrive[0] = _T('\0');
  TCHAR szDir[_MAX_DIR];
  szDir[0] = _T('\0');
  TCHAR szFname[_MAX_FNAME];
  szFname[0] = _T('\0');
  _tsplitpath_s(szIni, szDrive, sizeof(szDrive)/sizeof(TCHAR), szDir, sizeof(szDir)/sizeof(TCHAR), szFname, sizeof(szFname)/sizeof(TCHAR), nullptr, 0);
  _tmakepath_s(szIni, sizeof(szIni)/sizeof(TCHAR), szDrive, szDir, szFname, _T("ini"));
  free(const_cast<void*>(reinterpret_cast<const void*>(m_pszProfileName)));
  m_pszProfileName = _tcsdup(szIni);

  //Parse the command line and change the ini file if a filename is found
  CCommandLineInfo cmdInfo;
  ParseCommandLine(cmdInfo);
  if (!cmdInfo.m_strFileName.IsEmpty())
  {
    free(const_cast<void*>(reinterpret_cast<const void*>(m_pszProfileName)));
    m_pszProfileName = _tcsdup(cmdInfo.m_strFileName);
  }

  //setup all the virtual directories
  CSocksServerSettings settings;

  //Pull in all the configuration settings
  settings.m_nPort                    = static_cast<unsigned short>(GetProfileInt(_T("General"), _T("Port"), 1080));
  settings.m_sListenBindAddress       = GetProfileString(_T("General"), _T("ListenBindAddress"), _T(""));
  settings.m_sConnectBindAddress      = GetProfileString(_T("General"), _T("ConnectBindAddress"), _T(""));
  settings.m_dwReadTimeout            = GetProfileInt(_T("General"), _T("ReadTimeout"), settings.m_dwReadTimeout);
  settings.m_nThreadPoolSize          = GetProfileInt(_T("General"), _T("ThreadPoolSize"), settings.m_nThreadPoolSize);
  settings.m_dwWriteTimeout           = GetProfileInt(_T("General"), _T("WriteTimeout"), settings.m_dwWriteTimeout);
  settings.m_bEnableThreadLifetime    = GetProfileInt(_T("General"), _T("EnableThreadLifetime"), settings.m_bEnableThreadLifetime) != 0;
  settings.m_dwThreadLifetime         = GetProfileInt(_T("General"), _T("ThreadLifetime"), settings.m_dwThreadLifetime);
  bool bUseIOCPQueue                  = GetProfileInt(_T("General"), _T("UseIOCPQueue"), FALSE) != 0;
  settings.m_bSupportSocks4           = GetProfileInt(_T("General"), _T("Socks4"), settings.m_bSupportSocks4) != 0;
  settings.m_bSupportSocks5           = GetProfileInt(_T("General"), _T("Socks5"), settings.m_bSupportSocks5) != 0;
  settings.m_bSupportAnonymous        = GetProfileInt(_T("General"), _T("AnonymousAuthentication"), settings.m_bSupportAnonymous) != 0;
  settings.m_bSupportUserNamePassword = GetProfileInt(_T("General"), _T("UsernamePasswordAuthentication"), settings.m_bSupportUserNamePassword) != 0;
  settings.m_sUsername                = GetProfileString(_T("General"), _T("Username"), settings.m_sUsername);
  settings.m_sPassword                = GetProfileString(_T("General"), _T("Password"), settings.m_sPassword);

  if (!settings.m_bSupportSocks4 && !settings.m_bSupportSocks5)
  {
    _tprintf(_T("Socks Server must be configured to support Socks 4 and / or Socks 5\n"));
    WSACleanup();
    return FALSE;
  }

  if (!settings.m_bSupportAnonymous && !settings.m_bSupportUserNamePassword)
  {
    _tprintf(_T("Socks Server must be configured to support Anonymous and / or UserName / Password Authentication\n"));
    WSACleanup();
    return FALSE;
  }

  //and start it up
  _tprintf(_T("Socks server is starting...\n"));
  CMySocksServerIOCP theSocksServerIOCP;
  CMySocksServerDirected theSocksServerDirected;
  g_pTheSocksServerIOCP = &theSocksServerIOCP;
  g_pTheSocksServerDirected = &theSocksServerDirected;
  if (bUseIOCPQueue ? !theSocksServerIOCP.Start(&settings) : !theSocksServerDirected.Start(&settings))
  {
    _tprintf(_T("Failed to start up the socks server\n"));
    WSACleanup();
    return FALSE;
  }

  //Display some additional info when the web servering is starting
  _tprintf(_T("Listening for incoming connections on port %d\n"), static_cast<int>(settings.m_nPort));
  if (settings.m_sListenBindAddress.GetLength())
    _tprintf(_T("Binding to local address for incoming connections: %s\n"), settings.m_sListenBindAddress.GetString());
  if (settings.m_sConnectBindAddress.GetLength())
    _tprintf(_T("Binding to local address for onwards connections: %s\n"), settings.m_sConnectBindAddress.GetString());

  _tprintf(_T("Read timeout: %u ms\n"), settings.m_dwReadTimeout);
  _tprintf(_T("Write timeout: %u ms\n"), settings.m_dwWriteTimeout);
  _tprintf(_T("Thread pool size: %d\n"), settings.m_nThreadPoolSize);
  if (settings.m_bEnableThreadLifetime)
    _tprintf(_T("Threads in the thread pool will be recycled every %u minute(s)\n"), settings.m_dwThreadLifetime);
  if (bUseIOCPQueue)
    _tprintf(_T("Using IOCP Thread pool queue\n"));
  else
    _tprintf(_T("Using Directed Thread pool queue\n"));
  if (settings.m_bSupportSocks4)
    _tprintf(_T("Supporting Socks 4 clients\n"));
  if (settings.m_bSupportSocks5)
    _tprintf(_T("Supporting Socks 5 clients\n"));
  if (settings.m_bSupportAnonymous)
    _tprintf(_T("Supporting anonymous authentication\n"));
  if (settings.m_bSupportUserNamePassword)
  {
    _tprintf(_T("Supporting Username / Password authentication\n"));
    _tprintf(_T("Valid user is: %s\n"), settings.m_sUsername.GetString());
    _tprintf(_T("Valid password is: "));
    for (int i=0; i<settings.m_sPassword.GetLength(); i++)
      _tprintf(_T("*"));
    _tprintf(_T("\n"));   
  }

  //Check that everything in the thread pool also started up ok
  if (bUseIOCPQueue ? !theSocksServerIOCP.GetThreadPool().WaitForThreadsInitInstance() : !theSocksServerDirected.GetThreadPool().WaitForThreadsInitInstance())
  {
    WSACleanup();
    return FALSE;
  }

  //Register the console handler to allow the program to be gracefully terminated
  if (!SetConsoleCtrlHandler(ConsoleHandlerRoutine, TRUE))
  {
    WSACleanup();
    return FALSE;
  }

  //Wait until the server finishes
  if (bUseIOCPQueue)
    theSocksServerIOCP.Wait();
  else
    theSocksServerDirected.Wait();
  _tprintf(_T("Socks server has shut down...\n"));

  //Close down sockets
  WSACleanup();

  return FALSE;
}
